/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.paren;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.List;

import com.rapidminer.Process;
import com.rapidminer.example.Attribute;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.table.AttributeFactory;
import com.rapidminer.operator.ExecutionUnit;
import com.rapidminer.operator.IOContainer;
import com.rapidminer.operator.IOObject;
import com.rapidminer.operator.Model;
import com.rapidminer.operator.OperatorChain;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.meta.ParameterSet;
import com.rapidminer.operator.performance.PerformanceVector;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.preprocessing.filter.ChangeAttributeRole;
import com.rapidminer.operator.preprocessing.filter.attributes.AttributeFilter;
import com.rapidminer.operator.preprocessing.filter.attributes.SubsetAttributeFilter;
import com.rapidminer.operator.preprocessing.join.ExampleSetMerge;
import com.rapidminer.operator.tools.AttributeSubsetSelector;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeRepositoryLocation;
import com.rapidminer.repository.DataEntry;
import com.rapidminer.repository.Entry;
import com.rapidminer.repository.Folder;
import com.rapidminer.repository.IOObjectEntry;
import com.rapidminer.repository.RepositoryException;
import com.rapidminer.repository.RepositoryLocation;
import com.rapidminer.tools.Ontology;
import com.rapidminer.tools.OperatorService;
import com.rapidminer.tools.XMLException;

import de.dfki.madm.paren.casebase.CaseBase;
import de.dfki.madm.paren.casebase.RegressionModels;
import de.dfki.madm.paren.util.ConnectToInputPortExtender;

/**
 * This Operator takes a CaseBase as input. All datasets within the given CaseBase are
 * given once as input to the inner process. This process can be used to compute meta-features for each
 * dataset. The result of the inner process should be an exampleset with exactly one sample, which includes
 * the meta-features of the dataset. The computed meta-features of all datasets are collected within one
 * ExampleSet.
 * After this, the regression models for each classifier in the CaseBase are computed. This includes parameter
 * optimization for the LibSVM regression models (gamma, C), and computation of root mean squared errors.
 * The result is some CaseBaseRegressionModels object, which contains all necessary data for further
 * accuracy predictions on new data-sets.  
 *
 * @author Alexander Arimond
 *
 */

public class MetaFeaturesOperator extends OperatorChain {
	
	private static final String OPTIMIZATION_PROCESS_LOCATION = "com/rapidminer/resources/data/optimizationProcess.rmp";
	
	private static final String INPUTPORT_NAME = "dataset";
	private static final String OUTPUTPORT_NAME = "metafeatures";

	private static final String PARAMETER_DATASETS_LOCATION = "Path to datasets";
	
//	private static final String INPUTPORT_NAME = "in 1";
//	private static final String OUTPUTPORT_NAME = "out 1";
	
	private InputPort input = getInputPorts().createPort("casebase", CaseBase.class);
	
	private OutputPort innerSource = getSubprocess(0).getInnerSources().createPort(INPUTPORT_NAME);
	private InputPort innerSink = getSubprocess(0).getInnerSinks().createPort(OUTPUTPORT_NAME, ExampleSet.class);

	private OutputPort output = getOutputPorts().createPort("output");

	@Override
	public ExecutionUnit getSubprocess(int index) {
		// TODO Auto-generated method stub
		return super.getSubprocess(index);
	}
	
	public MetaFeaturesOperator(OperatorDescription description) {
		super(description, "Compute Meta-Features");
		getTransformer().addGenerationRule(innerSource, ExampleSet.class);
		getTransformer().addGenerationRule(output, RegressionModels.class);
		//getTransformer().addGenerationRule(output, ExampleSet.class);

	}
	
	// see JavaDocs of the Operator for documentation.
	// see inner comments for more detailed documentation.
	public void doWork() throws OperatorException{
		CaseBase casebase = input.getData();
		ExampleSet casebaseExampleSet = casebase.getCasebaseExampleSet();
		final RegressionModels models = new RegressionModels(casebase);
		
		// extract inner process for Meta-Features computation in order to recall it in further predicton steps:
		// for example in the PaREn classifier recommender wizard
		Process metaFeatureComputationProcess = extractInnerProcess();
		models.setMetaFeatureComputationProcess(metaFeatureComputationProcess);
		
		//System.out.println(metaFeatureComputationProcess.getRootOperator().getXML(false));
		
		try {
			ExampleSet metaFeatures = computeMetaFeatures(casebaseExampleSet);
			models.setMetaFeatures(metaFeatures);
			// optimize parameters
			// & compute regression models
				// load process - this is a standard process stored within the plugin
			Process optimizeProcess = loadOptimizeProcess();
				// for every classifier:
			List <String> classifierNames = casebase.getClassifierNames();
			for (String classifierName : classifierNames){
					// change role of classifier to label				
				ChangeAttributeRole changeRole = (ChangeAttributeRole) optimizeProcess.getOperator("Set Role");
				changeRole.setParameter(ChangeAttributeRole.PARAMETER_NAME, classifierName + " accuracy");
				
					// change selected attributes inside
				AttributeFilter selection = (AttributeFilter)optimizeProcess.getOperator("Select Attributes");
				//selection.setParameter(, value)
				selection.setParameter(AttributeSubsetSelector.PARAMETER_FILTER_TYPE, AttributeSubsetSelector.CONDITION_NAMES[2]); // this if for filter type: "subset"
				selection.setParameter(SubsetAttributeFilter.PARAMETER_ATTRIBUTES, getMetaFeaturesSelectionString(models));
					// run process
				System.out.println("regression modelling ... " + classifierName);

				IOContainer output = optimizeProcess.run(new IOContainer(models.getJoinedCaseBase()));
					// read optimized parameters
				ParameterSet params = output.get(ParameterSet.class);
					// read regression models with best parameters
				Model regressionModel = output.get(Model.class);
					// read root mean squared error
				PerformanceVector performance = output.get(PerformanceVector.class);

				models.addModel(classifierName, regressionModel, params, performance);	
			}
			output.deliver(models);
		} catch (RepositoryException e) {
			throw new OperatorException("repository exception", e);
		} catch (OperatorCreationException e) {
			throw new OperatorException("Could not create CaseBaseOperator", e);
		} catch (FileNotFoundException e) {
			throw new OperatorException("File not found", e);
		} catch (IOException e) {
			throw new OperatorException("Io exception", e);
		} catch (XMLException e) {
			throw new OperatorException("xml excpetion while reading process file", e);
		}

	}

	// this copies the inner process of the MetaFeaturesOperator into a seperate process
	// its quite hacky because for some reason it was not possible to name the inner sources and sinks the same as in the MetaFeaturesOperator
	private Process extractInnerProcess() {
		Process metaFeatureComputationProcess = new Process();
		ExecutionUnit unit = metaFeatureComputationProcess.getRootOperator().getSubprocess(0);

		String firstSinkOperatorName = getSubprocess(0).getInnerSources().getPortByIndex(0).getDestination().getPorts().getOwner().getOperator().getName();
		String firstSinkInputPortName = getSubprocess(0).getInnerSources().getPortByIndex(0).getDestination().getName();
		String lastSourceOperatorName = getSubprocess(0).getInnerSinks().getPortByIndex(0).getSource().getPorts().getOwner().getOperator().getName();
		String lastSourceOutputPortName = getSubprocess(0).getInnerSinks().getPortByIndex(0).getSource().getName();
		
		unit.getInnerSources().createPort(INPUTPORT_NAME);
		unit.getInnerSinks().createPort(OUTPUTPORT_NAME, ExampleSet.class);
		unit.cloneExecutionUnitFrom(getSubprocess(0), false);
		
		// connect inner sources and sinks with corresponding operators, dont know why unit.cloneExecutionUnitFrom(..) did not do that.
		unit.getInnerSources().getPortByName(INPUTPORT_NAME).disconnect();
		unit.getInnerSources().getPortByName("input 1").connectTo(unit.getOperatorByName(firstSinkOperatorName).getInputPorts().getPortByName(firstSinkInputPortName));
		unit.getOperatorByName(lastSourceOperatorName).getOutputPorts().getPortByName(lastSourceOutputPortName).disconnect();
		unit.getOperatorByName(lastSourceOperatorName).getOutputPorts().getPortByName(lastSourceOutputPortName).connectTo(unit.getInnerSinks().getPortByName("result 1"));
		return metaFeatureComputationProcess;
	}

	// create the selection string for the AttributeFilter Operator. The string is given as Parameter to the Operator
	private String getMetaFeaturesSelectionString(RegressionModels models) {
		List <String> metaFeatureNames = models.getMetaFeatureNames();
		StringBuilder builder = new StringBuilder();
		for (int i = 0; i < metaFeatureNames.size(); i++){
			builder.append(metaFeatureNames.get(i));
			if (i+1 != metaFeatureNames.size())
				builder.append("|");
		}
		System.out.println("Selected MetaFeatures for regression:" + builder.substring(0,builder.length()));
		return builder.substring(0, builder.length());
	}

	// load preconfigured optimizing process from plugin jar file 
	private Process loadOptimizeProcess() throws FileNotFoundException, IOException, XMLException {
		return new Process (readContent(this.getClass().getClassLoader().getResourceAsStream(OPTIMIZATION_PROCESS_LOCATION)));
	}

	// the inner process is executed here for every exampleSet used for building the CaseBase.
	// this inner process then computes the meta features in the way defined by the user
	// the results are collected in the end in one ExampleSet, each sample with ID attribute: data set path
	private ExampleSet computeMetaFeatures(ExampleSet casebaseExampleSet) throws OperatorCreationException, RepositoryException, OperatorException {
		
		RepositoryLocation location = getParameterAsRepositoryLocation(PARAMETER_DATASETS_LOCATION);
	    Entry pathEntry = location.locateEntry();
		
	    if (!(pathEntry instanceof Folder)){
	    	throw new OperatorException("You must specify a folder for your datasets");
	    }
	    
		ExampleSetMerge merger = OperatorService.createOperator(ExampleSetMerge.class);
		ConnectToInputPortExtender connector = new ConnectToInputPortExtender(merger);
		
		for (Example example : casebaseExampleSet){
			String dataSetName = example.getNominalValue(casebaseExampleSet.getAttributes().get(CaseBase.ATTRIBUTE_DATASET));
			// FIXME check if null is ok 
			IOObject object = getDEntry(dataSetName, (Folder) pathEntry).retrieveData(null);
			ExampleSet dataSet = (ExampleSet) object;
			innerSource.deliver(dataSet);
			getSubprocess(0).execute();
			ExampleSet resultSet = innerSink.getData();

			// set ID attribute: dataSetName
	        final Attribute idAttribute = AttributeFactory.createAttribute(CaseBase.ATTRIBUTE_DATASET, Ontology.NOMINAL);
			resultSet.getExampleTable().addAttribute(idAttribute);
			resultSet.getAttributes().setId(idAttribute);
			resultSet.getExampleTable().getDataRow(0).set(idAttribute, idAttribute.getMapping().mapString(dataSetName));

			InputPort connectedInputPort = connector.connectToNextInputPort();
			connectedInputPort.receive(resultSet);
		}

		merger.execute();
		ExampleSet finalResultSet = (ExampleSet) merger.getOutputPorts().getPortByIndex(0).getAnyDataOrNull();		
		return finalResultSet;
	}
	
    public static IOObjectEntry getDEntry(String name, Folder dataSetFolder) throws RepositoryException {
    	for(DataEntry entry:dataSetFolder.getDataEntries()) {
            if(entry.getName().equals(name)) {
                return (IOObjectEntry) entry;
            }
        }
        return null;
    }
	
	
	// just for reading the process.
    private  String readContent(InputStream inputStream) throws IOException {
        String content = "";
        BufferedReader reader = new BufferedReader(new InputStreamReader(inputStream));
        String line;
        while((line = reader.readLine()) != null) {
            content += line;
        }
        reader.close();
        return content;
    }
    
    @Override
    public List<ParameterType> getParameterTypes() {
        List<ParameterType> types = super.getParameterTypes();
        types.add(new ParameterTypeRepositoryLocation(PARAMETER_DATASETS_LOCATION, "Path to datasets", false));
        return types;
    }

}
