/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.paren;

import java.util.HashMap;
import java.util.Map.Entry;

import com.rapidminer.Process;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.ExampleSetFactory;
import com.rapidminer.operator.IOContainer;
import com.rapidminer.operator.Model;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.performance.PerformanceVector;
import com.rapidminer.operator.preprocessing.IdTagging;
import com.rapidminer.operator.preprocessing.filter.MissingValueReplenishment;
import com.rapidminer.tools.OperatorService;
import com.rapidminer.tools.container.Pair;
import common.Logger;

import de.dfki.madm.paren.landmarking.LandmarkingOperator;

public class Regressioner {

	public final static String ATTRIBUTE_PREDICTION = "prediction";
	public final static String ATTRIBUTE_RMSE = "rmse";
	public final static String ATTRIBUTE_CLASSIFIER = "classifier";
	
	private final static Logger logger = Logger.getLogger(Regressioner.class);
	
	private ExampleSet featureSet;
	private Process metaFeatureComputationProcess;


	
	public ExampleSet getFeatureSet() {
		return featureSet;
	}

	public void setMetaFeatureComputationProcess(Process process){
		this.metaFeatureComputationProcess = process;
	}
	
	public ExampleSet apply(ExampleSet dataSet, HashMap<String, Pair<Model, PerformanceVector>> models) throws OperatorException {
		try {
			// -- create meta features --
			if (this.metaFeatureComputationProcess == null){
				logger.info("No process given for meta feature computation, using simple LandmarkingOperator");
				featureSet = OperatorService.createOperator(LandmarkingOperator.class).doLandmarking(dataSet);
			} else {
				IOContainer result = this.metaFeatureComputationProcess.run(new IOContainer(dataSet));
				featureSet = result.get(ExampleSet.class);
			}
			
			MissingValueReplenishment missingValueReplenishment = OperatorService.createOperator(MissingValueReplenishment.class);
			missingValueReplenishment.setParameter(MissingValueReplenishment.PARAMETER_DEFAULT, "value");
			missingValueReplenishment.setParameter(MissingValueReplenishment.PARAMETER_REPLENISHMENT_VALUE, "0.5");
			Model pModel = missingValueReplenishment.createPreprocessingModel(featureSet);
			featureSet = pModel.apply(featureSet);
			// -- retrieve models from repository and apply them --
			HashMap<String, Pair<Double, Double>> predictions = new HashMap<String, Pair<Double,Double>>();
//			for(IOObjectCollection<IOObject> ioObjects:models.getObjects()) {
		    for(Entry<String, Pair<Model, PerformanceVector>> entry:models.entrySet()) {
			    String name = entry.getKey();
			    Model model = entry.getValue().getFirst();
			    PerformanceVector performanceVector = entry.getValue().getSecond();
				ExampleSet prediction = model.apply(featureSet);
				double predict = prediction.getExample(0).getNumericalValue(prediction.getAttributes().getPredictedLabel());
				double rmse = performanceVector.getCriterion("root_mean_squared_error").getFitness() * -1.0;
				predictions.put(name, new Pair<Double, Double>(predict, rmse));
			}
			// -- create example set of results --
			Object[][] result = new Object[predictions.size()][3];
			int i=0;
			for(Entry<String, Pair<Double, Double>> entry:predictions.entrySet()) {
				result[i][0] = entry.getKey();
				result[i][1] = entry.getValue().getFirst();
				result[i][2] = entry.getValue().getSecond();
				i++;
			}
			ExampleSet e1 = ExampleSetFactory.createExampleSet(result);
			e1.getAttributes().get("att1").setName(ATTRIBUTE_CLASSIFIER);
			e1.getAttributes().get("att2").setName(ATTRIBUTE_PREDICTION);
			e1.getAttributes().get("att3").setName(ATTRIBUTE_RMSE);
			
			IdTagging idTagging = OperatorService.createOperator(IdTagging.class);
			return idTagging.apply(e1);			
		} catch (Exception e) {
			e.printStackTrace();
			throw new OperatorException(e.getMessage(), e.getCause());
		}
	}
	


}
