/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.paren.casebase;

import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;

import com.rapidminer.Process;
import com.rapidminer.example.Attribute;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.operator.AbstractIOObject;
import com.rapidminer.operator.Annotations;
import com.rapidminer.operator.Model;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.meta.ParameterSet;
import com.rapidminer.operator.performance.PerformanceVector;
import com.rapidminer.operator.preprocessing.join.ExampleSetJoin;
import com.rapidminer.tools.OperatorService;
import com.rapidminer.tools.XMLException;

/**
 * This encapsulates the information of meta-feature computation for a given CaseBase. 
 * 
 * - All meta-features for all datasets in the casebase
 * - Regression models for all classifiers with respect to the meta-features
 * 
 * @author Alexander Arimond
 *
 */

public class RegressionModels extends AbstractIOObject {

	/**
	 * 
	 */
	private static final long serialVersionUID = 4090022660640295799L;
	
	
	private CaseBase casebase;
	private ExampleSet metaFeatures;

	private HashMap<String, ParameterSet> optimizedSVMParams = new HashMap<String, ParameterSet> ();
	private HashMap<String, Model> regressionModels = new HashMap<String, Model>();
	private HashMap<String, PerformanceVector> performances = new HashMap<String, PerformanceVector>();
	
	private ExampleSet joinedCasebaseTable;

	private String metaFeatureComputationProcessXML;


	public RegressionModels(CaseBase casebase){
		this.casebase = casebase;
		// train casebase regression models here
	}

	public void setMetaFeatures(ExampleSet metaFeatures) throws OperatorCreationException, OperatorException{
		this.metaFeatures = metaFeatures;
		ExampleSetJoin joiner = OperatorService.createOperator(ExampleSetJoin.class);
		joiner.getInputPorts().getPortByIndex(0).receive(this.casebase.getCasebaseExampleSet());
		joiner.getInputPorts().getPortByIndex(1).receive(this.metaFeatures);
		joiner.execute();
		this.joinedCasebaseTable = (ExampleSet)joiner.getOutputPorts().getPortByIndex(0).getAnyDataOrNull();
	}
	
	public ExampleSet getJoinedCaseBase(){
		return this.joinedCasebaseTable;
	}

	public List<String> getMetaFeatureNames() {
		if (metaFeatures == null){
			return new LinkedList<String>();
		}
		
		LinkedList<String> metaFeatureNames = new LinkedList<String>();
		// the metafeatures are the regular attributes, Path to dataset is the ID attribute
		for (Attribute attr: metaFeatures.getAttributes()){
			if (!attr.getName().equals(CaseBase.ATTRIBUTE_DATASET)){
				metaFeatureNames.add(attr.getName());
			}
		}
		return metaFeatureNames;
	}

	public void addModel(String classifierName, Model regressionModel, ParameterSet params,
			PerformanceVector performance) {
		this.regressionModels.put(classifierName, regressionModel);
		this.optimizedSVMParams.put(classifierName, params);
		this.performances.put(classifierName, performance);
	}
	
	public Model getModel(String classifierName){
		return this.regressionModels.get(classifierName);
	}
	
	public ParameterSet getOptimizedSVMParams(String classiferName){
		return this.optimizedSVMParams.get(classiferName);
	}
	
	public PerformanceVector getPerformanceVector(String classifierName){
		return this.performances.get(classifierName);
	}

	public List<String> getClassifierNames() {
		return casebase.getClassifierNames();
	}
	
	public CaseBase getCaseBase(){
		return casebase;
	}

	public void setMetaFeatureComputationProcess(Process metaFeatureComputationProcess) {
		this.metaFeatureComputationProcessXML = metaFeatureComputationProcess.getRootOperator().getXML(false);
	}
	
	public Process getMetaFeatureComputationProcess(){
		try {
			return new Process(this.metaFeatureComputationProcessXML);
			// exceptions should not happen because process is stored in a consistent way 
		} catch (IOException e) {
			e.printStackTrace();
		} catch (XMLException e) {
			e.printStackTrace();
		}
		return null;
	}

	@Override
	public Annotations getAnnotations() {
		return new Annotations();
	}

}
