/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.paren.gui.wizard;

import javax.swing.JFrame;
import javax.swing.JOptionPane;

import com.rapidminer.example.ExampleSet;
import com.rapidminer.gui.RapidMinerGUI;
import com.rapidminer.gui.tools.dialogs.wizards.AbstractWizard;
import com.rapidminer.gui.tools.dialogs.wizards.dataimport.RepositoryLocationSelectionWizardStep;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.io.RepositorySource;
import com.rapidminer.repository.Entry;
import com.rapidminer.repository.RepositoryLocation;
import com.rapidminer.tools.OperatorService;

/**
 * This initializes the Wizard and starts with selection of a repository location.
 * Step 1: The user can specify the dataset for which he likes to construct a classification system.
 * After this, predictions of the classfication accuracies are done for several classification methods.
 * Step 2: The user can select one or more of these classifiers for more detailed evaluation, according
 * to the predicted accuracies and the root mean squared error.
 * Step 3: After evaluation has been performed (essentially a parameter grid search with nested x-validation),
 * the actual accuracies are presented. The user may select his favorite classifier. A complete classification
 * system is constructed within the process view, including pre-processing steps.
 * 
 * @author Matthias Reif
 */
public class AutomaticSystemConstructionWizard extends AbstractWizard {
    
    /**
	 * 
	 */
	private static final long serialVersionUID = -9216872460213893764L;

    public RepositoryLocationSelectionWizardStepOwn dataSetRepositoryLocationStep;//, caseBaseRepositoryLocationStep;
    public RegressionWizardStep regressionWizardStep;
    public EvaluateWizardStep evaluateWizardStep;

    public AutomaticSystemConstructionWizard(String key, Object... i18nArgs) {
        super(RapidMinerGUI.getMainFrame(), key, i18nArgs);
        
        // first let user specify dataset
        dataSetRepositoryLocationStep = new RepositoryLocationSelectionWizardStepOwn("data", "select_repository_location_read", this, null, null);
        // perform regression to make predictions on the given dataset
        regressionWizardStep = new RegressionWizardStep("regression", this);
        // do detailed evaluation on the chosen classifiers
        evaluateWizardStep = new EvaluateWizardStep("evaluate", regressionWizardStep, this);

        addStep(dataSetRepositoryLocationStep);
        addStep(regressionWizardStep);
        addStep(evaluateWizardStep);

        layoutDefault();
    }
    
    /**
     * @return location of the repository selected in the dataset repository location step
     */
    public String getDataSetRepositoryLocation() {
        return dataSetRepositoryLocationStep.getRepositoryLocation();
    }

    /**
     * reads the dataset selected in the dataset repository location step
     * @return the dataset read as examplset
     * @throws OperatorCreationException
     * @throws OperatorException
     */
    public ExampleSet readDataSet() throws OperatorCreationException, OperatorException {
        RepositorySource repositorySource = OperatorService.createOperator(RepositorySource.class);
        repositorySource.setParameter(RepositorySource.PARAMETER_REPOSITORY_ENTRY, dataSetRepositoryLocationStep.getRepositoryLocation());
        return (ExampleSet) repositorySource.read();
    }

    class RepositoryLocationSelectionWizardStepOwn extends RepositoryLocationSelectionWizardStep {
        private String type;
        public RepositoryLocationSelectionWizardStepOwn(String type, String key, AbstractWizard parent, RepositoryLocation location, String initialValue) {
            super(key, parent, location, initialValue);
            this.type = type;
        }
        @Override
        protected boolean canProceed() {
            try {
                String repositoryLocation = getRepositoryLocation();
                if(repositoryLocation.isEmpty()) {
                    return false;
                }
                Entry entry = new RepositoryLocation(getRepositoryLocation()).locateEntry();
                if(entry == null) {
                    return false;
                }
                return entry.getType().equals(type);
            } catch(Exception e) {
                e.printStackTrace();
                return false;
            }
        }
    }
    
    @Override
    public void dispose() {
        evaluateWizardStep.stopThreads();
        System.gc();
        super.dispose();
    }
    
    public void exception(Exception e) {
        e.printStackTrace();
        JOptionPane.showMessageDialog(new JFrame(), e.getMessage(), "Error", JOptionPane.ERROR_MESSAGE);
        dispose();
    }
}
