/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.paren.gui.wizard;

import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.image.BufferedImage;
import java.util.HashMap;

import javax.swing.AbstractButton;
import javax.swing.ButtonGroup;
import javax.swing.DefaultCellEditor;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JProgressBar;
import javax.swing.JRadioButton;
import javax.swing.JTable;
import javax.swing.event.ChangeListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.DefaultTableCellRenderer;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.gui.tools.SwingTools;
import com.rapidminer.operator.performance.PerformanceVector;
import com.rapidminer.tools.I18N;
import com.rapidminer.tools.Tools;

import de.dfki.madm.paren.Regressioner;

/**
 * 
 * @author Matthias Reif
 *
 */

public class EvaluateTableModel extends AbstractTableModel implements ActionListener {

    /**
	 * 
	 */
	private static final long serialVersionUID = 8742646362200822187L;
	private String[] columnNames = { "Classifier", "Predicted Accuracy", "Root Mean Squared Error", "Actual Accuracy", "Cancel", "Load"};
    Attribute classifierAttribute, predictionAttribute, rmseAttribute;
    private HashMap<Double, Object> result;
    private HashMap<Double, JTableRadioButton> loadButtons;
    private HashMap<Double, JTableButton> cancelButtons;
    private HashMap<JTableButton, Double> cancelButtonsReverse;
    private HashMap<String, Double> ids;
    private ExampleSet predictions;
    private EvaluateWizardStep evaluateWizardStep;
    
    private static final int CANCEL_COLUMN = 4;
    private static final int LOAD_COLUMN = 5;

    public EvaluateTableModel(ExampleSet predictions, JTable table, HashMap<Double, Boolean> evaluate, EvaluateWizardStep evaluateWizardStep) {
        this.predictions = predictions;
        this.evaluateWizardStep = evaluateWizardStep;
        classifierAttribute = predictions.getAttributes().get(Regressioner.ATTRIBUTE_CLASSIFIER);
        predictionAttribute = predictions.getAttributes().get(Regressioner.ATTRIBUTE_PREDICTION);
        rmseAttribute = predictions.getAttributes().get(Regressioner.ATTRIBUTE_RMSE);
        result = new HashMap<Double, Object>();
        loadButtons = new HashMap<Double, JTableRadioButton>();
        cancelButtons = new HashMap<Double, JTableButton>();
        cancelButtonsReverse = new HashMap<JTableButton, Double>();
        ids = new HashMap<String, Double>();
        ButtonGroup buttonGroup = new ButtonGroup();
        for(Example example : predictions) {
            double id = example.getId();
            ids.put(example.getNominalValue(classifierAttribute), id);
            JTableRadioButton radioButton = new JTableRadioButton(table, LOAD_COLUMN, ids.size() - 1);
            radioButton.setBackground(Color.WHITE);
            buttonGroup.add(radioButton);
            loadButtons.put(id, radioButton);
            JTableButton cancelButton = new JTableButton(table, CANCEL_COLUMN, ids.size() - 1);
            cancelButton.addActionListener(this);
            cancelButtons.put(id, cancelButton);
            cancelButtonsReverse.put(cancelButton, id);
            if(evaluate.get(id)) {
                result.put(id, new JTableProgressBar(table, 3, ids.size() - 1));
            } else {
                result.put(id, "-");
                cancelButtons.get(id).setEnabled(false);
            }
        }
    }

    public void addChangeListener(ChangeListener listener) {
        for(JRadioButton radioButton : loadButtons.values()) {
            radioButton.addChangeListener(listener);
        }
    }

    public boolean isOneSelected() {
        for(Double id : loadButtons.keySet()) {
            if(loadButtons.get(id).isSelected()) {
                return true;
            }
        }
        return false;
    }

    public String getSelectedSystemName() {
        for(Example example : predictions) {
            double id = example.getId();
            String systemName = example.getNominalValue(classifierAttribute);
            if(loadButtons.get(id).isSelected()) {
                return systemName;
            }
        }
        return null;
    }

    public void setComputing(String systemName) {
        Object o = result.get(ids.get(systemName));
        if(o instanceof JTableProgressBar) {
//            ((ProgressRenderer) o).setIndeterminate(true);
            ((JTableProgressBar) o).setStringPainted(true);
        }
    }
    
    public void setProgress(String systemName, double progress) {
        Object o = result.get(ids.get(systemName));
        if(o instanceof JTableProgressBar) {
            ((JTableProgressBar) o).setValue((int)(progress * 100));
        }
    }

    public void setPerformance(String systemName, PerformanceVector performanceVector) {
        result.put(ids.get(systemName), Tools.formatNumber(performanceVector.getMainCriterion().getAverage()));
        cancelButtons.get(ids.get(systemName)).setEnabled(false);
        fireTableDataChanged();
    }

    public ExampleSet getPredictions() {
        return predictions;
    }

    @Override
    public int getColumnCount() {
        return columnNames.length;
    }

    @Override
    public int getRowCount() {
        return predictions.size();
    }

    @Override
    public String getColumnName(int col) {
        return columnNames[col];
    }

    @Override
    public Object getValueAt(int row, int col) {
        Example example = predictions.getExample(row);
        switch(col) {
        case 0:
            return example.getNominalValue(classifierAttribute);
        case 1:
            double prediction = example.getNumericalValue(predictionAttribute);
            if (prediction > 1.0) prediction = 1.0; // This may happen because of regression modelling, but looks strange to the user
            return Tools.formatNumber(prediction);
        case 2:
            return Tools.formatNumber(example.getNumericalValue(rmseAttribute));
        case 3:
            return result.get(example.getId());
        case CANCEL_COLUMN:
            return cancelButtons.get(example.getId());
        case LOAD_COLUMN:
            return loadButtons.get(example.getId());
        }
        return null;
    }

    public Class <? extends Object> getColumnClass(int col) {
        if(col >= 3) {
            return Object.class;
        } else {
            return String.class;
        }
    }

    public boolean isCellEditable(int row, int col) {
        return col == CANCEL_COLUMN || col == LOAD_COLUMN;
    }
    
    /**
     * cancel on one classifier evaluation clicked
     */
    @Override
    public void actionPerformed(ActionEvent e) {
        Double id = cancelButtonsReverse.get(e.getSource());
        evaluateWizardStep.cancelThread(cancelButtonsReverse.get(e.getSource()));
        result.put(id, "cancelled");
        cancelButtons.get(id).setEnabled(false);
        fireTableDataChanged();
    }
    
    class AbstractButtonEditor extends DefaultCellEditor implements ItemListener {
        /**
		 * 
		 */
		private static final long serialVersionUID = -8216786630949657785L;
		protected AbstractButton button;

        public AbstractButtonEditor(JCheckBox checkBox) {
            super(checkBox);
        }

        public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int row, int column) {
            if(value == null)
                return null;
            button = (AbstractButton) value;
            button.addItemListener(this);
            return (Component) value;
        }

        public Object getCellEditorValue() {
            button.removeItemListener(this);
            return button;
        }

        public void itemStateChanged(ItemEvent e) {
            super.fireEditingStopped();
        }
    }

    class JTableButton extends JButton {
        /**
		 * 
		 */
		private static final long serialVersionUID = 7950996735177291390L;
		private JTable table;
        private int column, row;

        public JTableButton(JTable table, int column, int row) {
            super();
            this.table = table;
            this.column = column;
            this.row = row;
            String iconLocation = I18N.getMessageOrNull(I18N.getGUIBundle(), "gui.action.cancel.icon");
            if (iconLocation != null) {
                ImageIcon smallIcon = SwingTools.createIcon("16/"+iconLocation);
                Image img = smallIcon.getImage();
                BufferedImage bi = new BufferedImage(12, 12, BufferedImage.TYPE_INT_ARGB);
                Graphics g = bi.createGraphics();
                g.drawImage(img, 2, 2, 8, 8, null);
                ImageIcon icon = new ImageIcon(bi);
                setIcon(icon);
            }
        }
        
        public void repaint() {
            // -- to trigger the cell repainting --
            if(table != null && table.getModel() != null) {
                ((AbstractTableModel) table.getModel()).fireTableCellUpdated(row, column);
            }
        }
    }
    
    class JTableRadioButton extends JRadioButton {
        /**
		 * 
		 */
		private static final long serialVersionUID = -7096238025090732549L;
		private JTable table;
        private int column, row;

        public JTableRadioButton(JTable table, int column, int row) {
            super();
            this.table = table;
            this.column = column;
            this.row = row;
        }
        
        public void repaint() {
            // -- to trigger the cell repainting --
            if(table != null && table.getModel() != null) {
                ((AbstractTableModel) table.getModel()).fireTableCellUpdated(row, column);
            }
        }
    }
    
    class JTableProgressBar extends JProgressBar {
        /**
		 * 
		 */
		private static final long serialVersionUID = -2038531846689892164L;
		private JTable table;
        private int column, row;

        public JTableProgressBar(JTable table, int column, int row) {
            super();
            this.table = table;
            this.column = column;
            this.row = row;
        }

        public boolean isDisplayable() {
            // -- to get the animation always running --
            return true;
        }

        public void repaint() {
            // -- to trigger the cell repainting --
            if(table != null && table.getModel() != null) {
                ((AbstractTableModel) table.getModel()).fireTableCellUpdated(row, column);
            }
        }
    }
    
    class JTableComponentRenderer extends DefaultTableCellRenderer {
        /**
		 * 
		 */
		private static final long serialVersionUID = -3435697265654945361L;

		@Override
        public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {
            if(value instanceof JComponent) {
                return (JComponent) value;
            } else {
                return super.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
            }
        }
    }
}
