/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.paren.gui.wizard;

import java.awt.event.WindowEvent;
import java.awt.event.WindowFocusListener;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import javax.swing.table.TableColumnModel;

import com.rapidminer.ConcurrencyTools;
import com.rapidminer.Process;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.gui.RapidMinerGUI;
import com.rapidminer.gui.tools.dialogs.wizards.WizardStep;
import com.rapidminer.operator.ExecutionUnit;
import com.rapidminer.operator.IOContainer;
import com.rapidminer.operator.MissingIOObjectException;
import com.rapidminer.operator.ModelApplier;
import com.rapidminer.operator.OperatorChain;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.ProcessStoppedException;
import com.rapidminer.operator.io.RepositorySource;
import com.rapidminer.operator.meta.ParameterIteratingOperatorChain;
import com.rapidminer.operator.meta.ParameterSet;
import com.rapidminer.operator.performance.PerformanceVector;
import com.rapidminer.operator.performance.PolynominalClassificationPerformanceEvaluator;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.validation.ParallelXValidation;
import com.rapidminer.operator.validation.XValidation;
import com.rapidminer.parameter.Parameters;
import com.rapidminer.tools.OperatorService;
import com.rapidminer.tools.RandomGenerator;
import com.rapidminer.tools.XMLException;

import de.dfki.madm.paren.GridSearchCount;
import de.dfki.madm.paren.GridSearchCount.Update;
import de.dfki.madm.paren.casebase.CaseBase;

/**
 * This is the third step in the Automatic System Construction Wizard.
 * The selected classifiers are evaluated by parameter grid search and x-validation.
 * In the end, one selected classifier will be used as basis for the construction of a 
 * classification process.
 * 
 * @author Matthias Reif, Alexander Arimond
 *
 */

class EvaluateWizardStep extends WizardStep implements WindowFocusListener {

    private RegressionWizardStep regressionWizardStep;
    private JPanel mainPanel;
    private AutomaticSystemConstructionWizard parent;
    private EvaluateTableModel model;
    private Map<String, ParameterSet> results;
    private ExecutorService threads;
    private HashMap<Double, EvaluateThread> workers;

    public EvaluateWizardStep(String i18nKey, RegressionWizardStep regressionWizardStep, AutomaticSystemConstructionWizard parent) {
        super(i18nKey);
        this.regressionWizardStep = regressionWizardStep;
        this.parent = parent;
        results = Collections.synchronizedMap(new HashMap<String, ParameterSet>());
        workers = new HashMap<Double, EvaluateThread>();
        mainPanel = new JPanel();
        mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
    }

    @Override
    protected boolean canGoBack() {
        return false;
    }

    @Override
    protected boolean canProceed() {
        return model != null && model.isOneSelected();
    }

    @Override
    protected JComponent getComponent() {
        return mainPanel;
    }

    @Override
    public boolean performEnteringAction() {
        initDialog();
        threads = Executors.newFixedThreadPool(regressionWizardStep.getNThreads());
        workers.clear();
        ClassifierTableModel classifierTableModel = regressionWizardStep.getModel();
        ExampleSet predictions = classifierTableModel.getPredictions();
        HashMap<Double, Boolean> evaluate = classifierTableModel.getEvaluations();
        for(Example example : predictions) {
            if(evaluate.get(example.getId())) {
                try {
                    String systemName = example.getNominalValue(classifierTableModel.classifierAttribute);
                    int seed = (int) new Date().getTime();
                    int folds = 4; // -- TODO: maybe configurable --
                    EvaluateThread evaluateThread = new EvaluateThread(systemName);
                    evaluateThread.setProcess(createOptimizeProcess(parent.regressionWizardStep.caseBaseRegressionModels.getCaseBase(), systemName, seed, folds, evaluateThread));
                    workers.put(example.getId(), evaluateThread);
                    threads.execute(evaluateThread);
                } catch(Exception e) {
                    parent.exception(e);
                }
            }
        }
        return true;
    }

    @Override
    public boolean performLeavingAction() {
        String systemName = model.getSelectedSystemName();
        if(systemName == null) {
            return false;
        }
        ParameterSet parameterSet = null;
        if(!results.containsKey(systemName)) {
            String msg = "The selected classifier was not optimized. Do you want to create an unoptimized process?";
            if(JOptionPane.showConfirmDialog(parent, msg, "Confirmation", JOptionPane.YES_NO_OPTION) == JOptionPane.NO_OPTION) {
                return false;
            }
        } else {
            parameterSet = results.get(systemName);
        }

        try {
            Process process = createProcess(this.parent.regressionWizardStep.caseBaseRegressionModels.getCaseBase(), systemName, parameterSet);
            RepositorySource repositorySource = OperatorService.createOperator(RepositorySource.class);
            repositorySource.setParameter(RepositorySource.PARAMETER_REPOSITORY_ENTRY, parent.getDataSetRepositoryLocation());
            process.getRootOperator().getSubprocess(0).addOperator(repositorySource);
            OutputPort p = process.getRootOperator().getSubprocess(0).getInnerSources().getPortByIndex(0);
            InputPort x = p.getDestination();
            p.disconnect();
            repositorySource.getOutputPorts().getPortByIndex(0).connectTo(x);
            RapidMinerGUI.getMainFrame().addWindowFocusListener(this);
            RapidMinerGUI.getMainFrame().setProcess(process, true);
        } catch(Exception e) {
            parent.exception(e);
            return false;
        }
        return true;
    }

	
    
    void initDialog() {
        JTable table = new JTable();

        ExampleSet predictions = regressionWizardStep.getModel().getPredictions();
        HashMap<Double, Boolean> evaluate = regressionWizardStep.getModel().getEvaluations();
        model = new EvaluateTableModel(predictions, table, evaluate, this);
        model.addChangeListener(parent);
        table.setModel(model);

        mainPanel.removeAll();
        mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));
        mainPanel.add(RegressionWizardStep.createResultPanel(table, regressionWizardStep.getDataSet(), regressionWizardStep));


        TableColumnModel dstTCM = table.getColumnModel();
        TableColumnModel srcTCM = regressionWizardStep.getTable().getColumnModel();
        for(int i = 0; i < 4; i++) {
            dstTCM.getColumn(i).setPreferredWidth(srcTCM.getColumn(i).getWidth());
        }
        table.getColumn("Actual Accuracy").setCellRenderer(model.new JTableComponentRenderer());
        
        table.getColumn("Load").setCellRenderer(model.new JTableComponentRenderer());
        table.getColumn("Load").setCellEditor(model.new AbstractButtonEditor(new JCheckBox()));
       
        table.getColumn("Cancel").setCellRenderer(model.new JTableComponentRenderer());
        table.getColumn("Cancel").setCellEditor(model.new AbstractButtonEditor(new JCheckBox()));
    }

    public class EvaluateThread extends SwingWorker<IOContainer, Object> implements Update {
        private String systemName;
        private Process process;

        public EvaluateThread(String systemName) {
            this.systemName = systemName;
        }
        
        public void setProcess(Process process) {
        	//System.out.println(process.getRootOperator().getXML(false));
            this.process = process;
        }

        public void stop() {
            if(process != null) {
                process.stop();
            }
        }

        @Override
        public IOContainer doInBackground() {
            if(process == null) {
                return null;
            }
            try {
                SwingUtilities.invokeLater(new Runnable() {
                    @Override
                    public void run() {
                        model.setComputing(systemName);
                    }
                });
                ExampleSet dataSetCopy = null;
                synchronized(parent) {
                    dataSetCopy = parent.readDataSet();
                }
                return process.run(new IOContainer(dataSetCopy));
            } catch(ProcessStoppedException e1) {
            } catch(OperatorCreationException e) {
                parent.exception(e);
            } catch(OperatorException e) {
                parent.exception(e);
            }
            return null;
        }

        @Override
        protected void done() {
            try {
                if(!isCancelled()) {
                    IOContainer result = get();
                    if(result != null) {
                        PerformanceVector performanceVector = result.get(PerformanceVector.class);
                        ParameterSet parameterSet = null;
                        if(result.contains(ParameterSet.class)) {
                            parameterSet = result.get(ParameterSet.class);
                        }
                        synchronized(model) {
                            model.setPerformance(systemName, performanceVector);
                        }
                        results.put(systemName, parameterSet);
                    }
                }
            } catch(InterruptedException e) {
                e.printStackTrace();
            } catch(ExecutionException e) {
                parent.exception(e);
            } catch(MissingIOObjectException e) {
                parent.exception(e);
            }
        }

        @Override
        public void update(double value) {
            final double x = value; 
            SwingUtilities.invokeLater(new Runnable() {
                @Override
                public void run() {
                    model.setProgress(systemName, x);
                }
            });
        }
    }
    
    void cancelThread(double id) {
        workers.get(id).stop();
        if(!workers.get(id).isDone()) {
            workers.get(id).cancel(true);
        }
    }

    @Override
    public void windowGainedFocus(WindowEvent arg0) {
        RapidMinerGUI.getMainFrame().getProcessPanel().getProcessRenderer().ARRANGE_OPERATORS_ACTION.actionPerformed(null);
    }

    @Override
    public void windowLostFocus(WindowEvent arg0) {
    }
    
    public void stopThreads() {
        for(EvaluateThread evaluateThread : workers.values()) {
            evaluateThread.stop();
        }
        for(EvaluateThread evaluateThread : workers.values()) {
            if(!evaluateThread.isDone()) {
                evaluateThread.cancel(true);
            }
        }
    }

    /**
     * creates the process which is given as constructed process in the end.
     */
    
    private static Process createProcess(CaseBase casebase, String classifier, ParameterSet parameterSet) throws IOException, XMLException {
        //String preprocessXML = casebase.getPreprocess(classifier).getRootOperator().getXML(false);
        //String processXML = casebase.getProcess(classifier).getRootOperator().getXML(false);
        
        Process preprocess = casebase.getPreprocess(classifier);
        Process process = casebase.getProcess(classifier);
        
        Process newProcess = new Process();
        OperatorChain root = newProcess.getRootOperator();
        OutputPort preprocessOutput = null;
        InputPort totalInput = null;
            
        if(preprocess != null) {
            ExecutionUnit preprocessUnit = preprocess.getRootOperator().getSubprocess(0);
            totalInput = preprocessUnit.getInnerSources().getPortByIndex(0).getDestination();
            preprocessOutput = preprocessUnit.getInnerSinks().getPortByIndex(0).getSource();
            root.getSubprocess(0).stealOperatorsFrom(preprocessUnit);
            root.getSubprocess(0).getInnerSources().disconnectAll();
            root.getSubprocess(0).getInnerSinks().disconnectAll();
        }
        
        ExecutionUnit processUnit = process.getRootOperator().getSubprocess(0);
        InputPort processInput = processUnit.getInnerSources().getPortByIndex(0).getDestination();
        root.getSubprocess(0).stealOperatorsFrom(processUnit);
        root.getSubprocess(0).getInnerSources().disconnectAll();

        if(preprocessOutput != null) {
            preprocessOutput.connectTo(processInput);
        } else {
            totalInput = processInput;
        }
        
        root.getSubprocess(0).getInnerSources().getPortByIndex(0).connectTo(totalInput);
        
        if(parameterSet != null) {
            parameterSet.applyAll(newProcess, new HashMap<String, String>());
        }
        return newProcess;
    }
    
    /**
     * This creates the optimization process for classifier evaluation. 
     * This process consists of parameter grid search operator.
     * Nested in this there are pre-processing (depending on classifier) steps and a X-Validation.
     * The classifier will be evaluated for different parameter settings via grid search.
     * The process delivers a performance vector with best accuracy and used parameters. 
     */
    private static Process createOptimizeProcess(CaseBase casebase, String classifier, int seed, int folds, Update update)
    throws IOException, XMLException, OperatorCreationException {
        
        Process preprocess = casebase.getPreprocess(classifier);
        Process process = casebase.getProcess(classifier);
        String params = casebase.getOptimizationParams(classifier).toString();
        
        // this is a HACK for windows compatibility .. - AA
        // Parameters are usually stored with to symbols: Parameters.PAIR_SEPARATOR = '\u241d' and Parameters.RECORD_SEPARATOR = '\u241e'
        // because on windows there is another standard encoding, i replaced all those characters by the words PAIR_SEPARATOR and RECORD_SEPARATOR
        // in the regressionModels.ioo file and here they are replaced back with the correctly encoded characters.
        // this should be reviewed in later versions of the wizard.
        // the old regressionModels.ioo can be found as regressionModels_OLD.ioo
        params = params.replaceAll("PAIR_SEPARATOR", new String(new char[]{Parameters.PAIR_SEPARATOR})).replaceAll("RECORD_SEPARATOR", new String(new char[] {Parameters.RECORD_SEPARATOR}));

        
        
        ExecutionUnit preprocessUnit = null;
        InputPort ipPre = null;
        OutputPort opPre = null;
        if(preprocess != null) {
            preprocessUnit = preprocess.getRootOperator().getSubprocess(0);
            ipPre = preprocessUnit.getInnerSources().getPortByIndex(0).getDestination();
            opPre = preprocessUnit.getInnerSinks().getPortByIndex(0).getSource();
            preprocessUnit.getInnerSources().disconnectAll();
            preprocessUnit.getInnerSinks().disconnectAll();
        }

        // -- create inner optimization chain --
        XValidation validation = OperatorService.createOperator(ParallelXValidation.class);
        ModelApplier modelApplier = OperatorService.createOperator(ModelApplier.class);
        PolynominalClassificationPerformanceEvaluator performanceEvaluator = OperatorService.createOperator(PolynominalClassificationPerformanceEvaluator.class);
        validation.getSubprocess(1).addOperator(modelApplier);
        validation.getSubprocess(1).getInnerSources().getPortByName("model").connectTo(modelApplier.getInputPorts().getPortByName("model"));
        validation.getSubprocess(1).getInnerSources().getPortByName("test set").connectTo(modelApplier.getInputPorts().getPortByName("unlabelled data"));
        validation.getSubprocess(1).addOperator(performanceEvaluator);
        validation.setParameter(XValidation.PARAMETER_NUMBER_OF_VALIDATIONS, Integer.toString(folds));
        validation.setParameter(RandomGenerator.PARAMETER_USE_LOCAL_RANDOM_SEED, Boolean.TRUE.toString());
        validation.setParameter(RandomGenerator.PARAMETER_LOCAL_RANDOM_SEED, Integer.toString(seed));
        validation.setParameter(ConcurrencyTools.PARAMETER_NUMBER_OF_THREADS, "4");
        modelApplier.getOutputPorts().getPortByName("labelled data").connectTo(performanceEvaluator.getInputPorts().getPortByName("labelled data"));
        performanceEvaluator.getOutputPorts().getPortByName("performance").connectTo(validation.getSubprocess(1).getInnerSinks().getPortByName("averagable 1"));
        performanceEvaluator.setParameter(PolynominalClassificationPerformanceEvaluator.PARAMETER_MAIN_CRITERION, "accuracy");
        Process optimizeProcess = new Process();
        ExecutionUnit optimizeUnit = optimizeProcess.getRootOperator().getSubprocess(0);
        // -- NO parameters for optimization set => just x-validation --
        if(params.isEmpty()) {
            // -- add operators to optimization --
            if(preprocessUnit != null) {
                optimizeUnit.stealOperatorsFrom(preprocessUnit);
            }
            optimizeUnit.addOperator(validation);
            // -- wire operators --
            if(preprocessUnit != null) {
                optimizeUnit.getInnerSources().getPortByIndex(0).connectTo(ipPre);
                opPre.connectTo(validation.getInputPorts().getPortByName("training"));
            } else {
                optimizeUnit.getInnerSources().getPortByIndex(0).connectTo(validation.getInputPorts().getPortByName("training"));
            }
            validation.getOutputPorts().getPortByName("averagable 1").connectTo(optimizeUnit.getInnerSinks().getPortByIndex(0));
            // -- parameters for optimization set => add parameter optimization operator --
        } else {
            System.out.println("Parameters for optimization: " + params);
            GridSearchCount optimizer = OperatorService.createOperator(GridSearchCount.class);
            optimizer.setUpdate(update);

            optimizeUnit.addOperator(optimizer);
            optimizer.setParameter(ParameterIteratingOperatorChain.PARAMETER_PARAMETERS, params);

            // -- add operators to optimization --
            if(preprocessUnit != null) {
                optimizer.getSubprocess(0).stealOperatorsFrom(preprocessUnit);
            }
            optimizer.getSubprocess(0).addOperator(validation);

            // -- wire operators --
            if(preprocessUnit != null) {
                optimizer.getSubprocess(0).getInnerSources().getPortByIndex(0).connectTo(ipPre);
                opPre.connectTo(validation.getInputPorts().getPortByName("training"));
            } else {
                optimizer.getSubprocess(0).getInnerSources().getPortByIndex(0).connectTo(validation.getInputPorts().getPortByName("training"));
            }
            validation.getOutputPorts().getPortByName("averagable 1").connectTo(optimizer.getSubprocess(0).getInnerSinks().getPortByIndex(0));
            
            optimizeUnit.getInnerSources().getPortByIndex(0).connectTo(optimizer.getInputPorts().getPortByIndex(0));
            optimizer.getOutputPorts().getPortByName("performance").connectTo(optimizeUnit.getInnerSinks().getPortByIndex(0));
            optimizer.getOutputPorts().getPortByName("parameter").connectTo(optimizeUnit.getInnerSinks().getPortByIndex(1));

        }

        // -- inject classifier --
        ExecutionUnit classifierUnit = process.getRootOperator().getSubprocess(0);
        InputPort ip = classifierUnit.getInnerSources().getPortByIndex(0).getDestination();
        OutputPort op = classifierUnit.getInnerSinks().getPortByIndex(0).getSource();
        // FIXME Weka dependency should be resolved by MANIFEST entry. but still this is HACK -- find a nicer solution. 
        if (op == null){
            throw new NullPointerException("Operator initialization failed. Probably you dont have Weka extension installed.");
        }
        validation.getSubprocess(0).stealOperatorsFrom(classifierUnit);
        validation.getSubprocess(0).getInnerSources().getPortByIndex(0).connectTo(ip);
        op.connectTo(validation.getSubprocess(0).getInnerSinks().getPortByIndex(0));
        return optimizeProcess;
    }
    
}
