/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.paren.gui.wizard;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JTable;
import javax.swing.RowSorter;
import javax.swing.SortOrder;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingWorker;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.Example;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.gui.tools.dialogs.wizards.WizardStep;
import com.rapidminer.operator.AbstractIOObject;
import com.rapidminer.operator.IOObject;
import com.rapidminer.operator.Model;
import com.rapidminer.operator.performance.PerformanceVector;
import com.rapidminer.parameter.Parameters;
import com.rapidminer.tools.Tools;
import com.rapidminer.tools.XMLSerialization;
import com.rapidminer.tools.container.Pair;
import com.rapidminer.tools.io.Encoding;

import de.dfki.madm.paren.Regressioner;
import de.dfki.madm.paren.casebase.RegressionModels;


/**
 * This is the second step of the PaREn Automatic System Construction Wizard.
 * Predictions on accuracies are made for several classifiers based on previous accuracies (casebase) 
 * and meta features (e.g. landmarking).
 * 
 * @author Matthias Reif, Alexander Arimond
 *
 */

class RegressionWizardStep extends WizardStep implements ChangeListener, ActionListener {

    // this contains all the statistical information for the peforming the Regression and compute the predictions 
	private static final String CASEBASE_REGRESSION_MODELS = "com/rapidminer/resources/data/regressionModels.ioo";
	
    private JPanel mainPanel;
    private JPanel progressPanel, resultPanel;
    private AutomaticSystemConstructionWizard parent;
    private JTable table;
    private ClassifierTableModel model;
    private ExampleSet dataSet, predictions, featureSet;
    private JSpinner nThreadsSpinner;
    private JButton selectAllButton;

    public RegressionModels caseBaseRegressionModels;

    private static final int N_THREADS_DEFAULT = 1;
    
    public RegressionWizardStep(String i18nKey, AutomaticSystemConstructionWizard parent) {
        super(i18nKey);
        this.parent = parent;
    }

    @Override
    protected boolean canGoBack() {
        return false;
    }

    @Override
    protected boolean canProceed() {
        return model != null;
    }

    public int getNThreads() {
        return (Integer) nThreadsSpinner.getValue();
    }
    
    public ClassifierTableModel getModel() {
        return model;
    }
    
    public JTable getTable() {
        return table;
    }
    
    public ExampleSet getDataSet() {
        return dataSet;
    }

    @Override
    protected JComponent getComponent() {
        mainPanel = new JPanel();
        mainPanel.setLayout(new BoxLayout(mainPanel, BoxLayout.Y_AXIS));

        progressPanel = new JPanel();
        progressPanel.setLayout(new BoxLayout(progressPanel, BoxLayout.PAGE_AXIS));

        JProgressBar progressBar = new JProgressBar();
        progressBar.setIndeterminate(true);

        progressPanel.add(Box.createRigidArea(new Dimension(20, 20)));
        progressPanel.add(new JLabel("Please wait while analyzing the data..."));
        progressPanel.add(Box.createRigidArea(new Dimension(20, 20)));
        progressPanel.add(progressBar);
        progressPanel.add(Box.createVerticalGlue());

        mainPanel.add(progressPanel);

        return mainPanel;
    }

    protected static JScrollPane getTablePanel(JTable table) {
        table.setAutoCreateRowSorter(true);
        ArrayList<RowSorter.SortKey> sortKeys = new ArrayList<RowSorter.SortKey>();
        sortKeys.add(new RowSorter.SortKey(1, SortOrder.DESCENDING));
        sortKeys.add(new RowSorter.SortKey(2, SortOrder.ASCENDING));
        sortKeys.add(new RowSorter.SortKey(0, SortOrder.ASCENDING));
        table.getRowSorter().setSortKeys(sortKeys);
        JScrollPane scrollPane = new JScrollPane(table);
        scrollPane.setBorder(BorderFactory.createLineBorder(Color.BLACK));
        table.setFillsViewportHeight(true);
        return scrollPane;
    }

    protected void createMainPanel(JComponent mainPanel) {
        model = new ClassifierTableModel(predictions);
        table = new JTable(model);
        progressPanel.setVisible(false);
        mainPanel.remove(progressPanel);
        parent.validate();
        resultPanel = createResultPanel(table, dataSet, this);

        SpinnerNumberModel spinnerModel = new SpinnerNumberModel(N_THREADS_DEFAULT, 1, Integer.MAX_VALUE, 1);
        
        selectAllButton = new JButton(new AbstractAction("Select All") {
            /**
			 * 
			 */
			private static final long serialVersionUID = -5430858335384041735L;

			@Override
            public void actionPerformed(ActionEvent e) {
                for(int i=0; i<model.getRowCount(); i++) {
                    model.setValueAt(true, i, 3);
                }
            }
        });
        Box buttonBox = new Box(BoxLayout.LINE_AXIS);
        buttonBox.add(Box.createHorizontalGlue());
        buttonBox.add(selectAllButton);
        buttonBox.add(Box.createHorizontalGlue());
        resultPanel.add(buttonBox, BorderLayout.PAGE_END);
        
        nThreadsSpinner = new JSpinner(spinnerModel);

        mainPanel.add(resultPanel);
        mainPanel.revalidate();
        parent.validate();
    }

    protected static JPanel createResultPanel(JTable table, ExampleSet dataSet, ActionListener actionListener) {
        String msg = "Dataset has " + dataSet.size() + " samples with " + dataSet.getAttributes().size() + " attributes.";
        Box topBox = Box.createHorizontalBox();
        topBox.add(new JLabel(msg));
        if(actionListener != null) {
            topBox.add(Box.createHorizontalStrut(20));
            ((JButton) topBox.add(new JButton("Show Meta Features"))).addActionListener(actionListener);
        }
        JPanel panel = new JPanel(new BorderLayout(0, 10));
        panel.add(topBox, BorderLayout.PAGE_START);
        panel.add(getTablePanel(table), BorderLayout.CENTER);
        return panel;
    }
    
    @Override
    protected boolean performEnteringAction() {
        new SwingWorker<ExampleSet[], Object>() {
            


            @Override
            public ExampleSet[] doInBackground() {
                try {
                    dataSet = parent.readDataSet();
                    
                    // load CaseBase and regression models, i.e. all statistical and other information for doing the 
                    // predictions by regression
                    // and performing the more detailed evaluation of classifiers afterwards
                    caseBaseRegressionModels = loadDefaultCaseBaseRegressionModels();
                    List<String> classifierNames = caseBaseRegressionModels.getClassifierNames();
                    
                    HashMap<String, Pair<Model, PerformanceVector>> models = new HashMap<String, Pair<Model, PerformanceVector>>();
                    for (String classifier : classifierNames){
                    	models.put(classifier, new Pair<Model, PerformanceVector>(caseBaseRegressionModels.getModel(classifier), caseBaseRegressionModels.getPerformanceVector(classifier)));
                    }
                    // compute predictions and feature set
                    Regressioner regressioner = new Regressioner();
                    regressioner.setMetaFeatureComputationProcess(caseBaseRegressionModels.getMetaFeatureComputationProcess());
                    ExampleSet predictions = regressioner.apply(dataSet, models);
                    ExampleSet featureSet = regressioner.getFeatureSet();
                    regressioner = null; // FIXME needed ?
                    return new ExampleSet[] { predictions, featureSet };
                } catch(Exception e) {
                    parent.exception(e);
                    return null;
                }
            }

            private RegressionModels loadDefaultCaseBaseRegressionModels() throws IOException {
            	URL url = this.getClass().getClassLoader().getResource(CASEBASE_REGRESSION_MODELS);
            	if (url == null){
            		throw new IOException("Could not load resource " + CASEBASE_REGRESSION_MODELS +".");
            	}
            	return (RegressionModels) readIOObjectFromResource(CASEBASE_REGRESSION_MODELS);
			}

			@Override
            protected void done() {
                try {
                    ExampleSet[] result = get();
                    predictions = result[0];
                    featureSet = result[1];
                    createMainPanel(mainPanel);
                    parent.stateChanged(new ChangeEvent(this));
                } catch(Exception e) {
                    e.printStackTrace();
                }
            }
        }.execute();
        return true;
    }

    @Override
    public void stateChanged(ChangeEvent arg0) {
    }

    public void showFeatureSet() {
        String[] columnNames = { "Feature", "Value" };
        Object[][] values = new Object[featureSet.getAttributes().size()][2];
        for(Example example : featureSet) {
            int i = 0;
            for(Attribute attribute : featureSet.getAttributes()) {
                values[i][0] = attribute.getName();
                if(attribute.isNominal()) {
                    values[i][1] = example.getNominalValue(attribute);
                } else {
                    values[i][1] = Tools.formatNumber(example.getNumericalValue(attribute));
                }
                i++;
            }
        }
        JTable table = new JTable(values, columnNames);
        JScrollPane scrollPane = new JScrollPane(table);
        scrollPane.setBorder(BorderFactory.createLineBorder(Color.BLACK));
        JOptionPane.showMessageDialog(parent, scrollPane);
    }

    private static IOObject readIOObjectFromResource(final String resource) throws IOException {
        return AbstractIOObject.read(new AbstractIOObject.InputStreamProvider() {
            @Override
            public InputStream getInputStream() throws IOException {
                
                return this.getClass().getClassLoader().getResourceAsStream(resource);
            }
        });
    }
    
    @Override
    public void actionPerformed(ActionEvent event) {
        showFeatureSet();
    }
}
