/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.paren.landmarking;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import com.rapidminer.example.ExampleSet;
import com.rapidminer.operator.ModelApplier;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorChain;
import com.rapidminer.operator.OperatorCreationException;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.performance.PerformanceEvaluator;
import com.rapidminer.operator.performance.PerformanceVector;
import com.rapidminer.operator.performance.PolynominalClassificationPerformanceEvaluator;
import com.rapidminer.operator.preprocessing.normalization.Normalization;
import com.rapidminer.operator.validation.XValidation;
import com.rapidminer.tools.LogService;
import com.rapidminer.tools.OperatorService;

/**
 * Abstract superclass of all landmark learner classes, used in the landmark
 * operator. Any landmark learner as a subclass of {@link AbstractLandmarker}
 * can be evaluated using Cross Validation depending on the value of the cross
 * validation parameter.
 * 
 * @author Sarah Daniel
 * 
 */
public abstract class AbstractLandmarker {

    protected String PARAMETER_PERFORMANCE_ACCURACY = "accuracy";

    protected PerformanceVector performance;
    protected String name;

    protected void learnExampleSet(ExampleSet exampleSet, Operator learner, Map<String, String> parameters) throws OperatorCreationException, OperatorException {
        if (Boolean.valueOf(parameters.get(LandmarkingParameters.PARAMETER_CROSS_VALIDATION)))
            xValidation(exampleSet, learner, parameters);
        else
            learn(exampleSet, learner, parameters);

    }

    /**
     * The {@link Operator} learner takes the {@link ExampleSet} as input and
     * then its model is applied to the {@link ExampleSet}. Finally the
     * performance is evaluated.
     * 
     * @param exampleSet
     *            the {@link ExampleSet} used as input for the {@link Operator}
     *            learner.
     * @param learner
     *            the {@link Operator} that should be evaluated as a landmark.
     * @param parameters
     *            the {@link Map} containing all parameters of the landmarking.
     */
    protected void learn(ExampleSet exampleSet, Operator learner, Map<String, String> parameters) {
        try {

            Operator modelApplier = OperatorService.createOperator(ModelApplier.class);
            Operator performanceEvaluator = createPerformanceEvaluator();

            learner.getInputPorts().getPortByIndex(0).receive(exampleSet);
            learner.getOutputPorts().getPortByIndex(0).connectTo(modelApplier.getInputPorts().getPortByIndex(0));
            learner.getOutputPorts().getPortByIndex(1).connectTo(modelApplier.getInputPorts().getPortByIndex(1));
            modelApplier.getOutputPorts().getPortByIndex(0).connectTo(performanceEvaluator.getInputPorts().getPortByIndex(0));

            learner.execute();
            modelApplier.execute();
            performanceEvaluator.execute();

            this.performance = performanceEvaluator.getOutputPorts().getPortByIndex(0).getData();

        } catch (OperatorCreationException e) {
            e.printStackTrace();
        } catch (OperatorException e) {
            e.printStackTrace();
        }
    }

    /**
     * An {@link XValidation} is applied on the {@link Operator} learner.
     * 
     * @param exampleSet
     *            the {@link ExampleSet} used as input for the
     *            {@link XValidation} process.
     * @param learner
     *            the {@link Operator} is an inner operator of the X-Validation
     *            process that should be evaluated as a landmark.
     * @param parameters
     *            the {@link Map} containing all parameters of the landmarking.
     * @throws OperatorCreationException 
     * @throws OperatorException 
     */
    protected void xValidation(ExampleSet exampleSet, Operator learner, Map<String, String> parameters) throws OperatorCreationException, OperatorException {
            OperatorChain xValidation = (OperatorChain) OperatorService.createOperator(XValidation.class);

            if (!(exampleSet.size() > Integer.valueOf(parameters.get(LandmarkingParameters.PARAMETER_NUMBER_OF_VALIDATIONS)))) {
                xValidation.setParameter(XValidation.PARAMETER_LEAVE_ONE_OUT, "true");
                LogService.getRoot().info(
                        "The ExampleSet has only " + exampleSet.size() + " examples. So a leave one out cross-validation will be performed instead of a "
                                + parameters.get(LandmarkingParameters.PARAMETER_NUMBER_OF_VALIDATIONS) + " fold cross-validatoin.");
            }

            xValidation.setParameter(XValidation.PARAMETER_NUMBER_OF_VALIDATIONS, parameters.get(LandmarkingParameters.PARAMETER_NUMBER_OF_VALIDATIONS));
            xValidation.setParameter(XValidation.PARAMETER_SAMPLING_TYPE, parameters.get(LandmarkingParameters.PARAMETER_SAMPLING_TYPE));
            xValidation.setParameter(XValidation.PARAMETER_AVERAGE_PERFORMANCES_ONLY, parameters.get(LandmarkingParameters.PARAMETER_AVERAGE_PERFORMANCES_ONLY));

            xValidation.getInputPorts().getPortByIndex(0).receive(exampleSet);

            Operator modelApplier = OperatorService.createOperator(ModelApplier.class);
            Operator performanceEvaluator = createPerformanceEvaluator();

            xValidation.getSubprocess(0).addOperator(learner);
            xValidation.getSubprocess(1).addOperator(modelApplier);
            xValidation.getSubprocess(1).addOperator(performanceEvaluator);

            // Connecting Ports
            xValidation.getSubprocess(0).getInnerSources().getPortByIndex(0).connectTo(learner.getInputPorts().getPortByIndex(0));
            learner.getOutputPorts().getPortByIndex(0).connectTo(xValidation.getSubprocess(0).getInnerSinks().getPortByIndex(0));

            xValidation.getSubprocess(1).getInnerSources().getPortByIndex(0).connectTo(modelApplier.getInputPorts().getPortByIndex(0));
            xValidation.getSubprocess(1).getInnerSources().getPortByIndex(1).connectTo(modelApplier.getInputPorts().getPortByIndex(1));

            modelApplier.getOutputPorts().getPortByIndex(0).connectTo(performanceEvaluator.getInputPorts().getPortByIndex(0));
            performanceEvaluator.getOutputPorts().getPortByIndex(0).connectTo(xValidation.getSubprocess(1).getInnerSinks().getPortByIndex(0));

            // System.out.println("Execute XValidation");
            xValidation.execute();

            this.performance = xValidation.getOutputPorts().getPortByIndex(2).getData();


    }

    protected ExampleSet normalizeExampleSet(ExampleSet exampleSet, Map<String, String> parameters) throws OperatorException, OperatorCreationException {


            if (Boolean.valueOf(parameters.get(LandmarkingParameters.PARAMETER_NORMALIZE_EXAMPLESET))) {
                Normalization normalizer = OperatorService.createOperator(Normalization.class);
                normalizer.setParameter(Normalization.PARAMETER_NORMALIZATION_METHOD, parameters.get(LandmarkingParameters.PARAMETER_NORMALIZATION_METHOD));
                normalizer.setParameter(Normalization.PARAMETER_MIN, parameters.get(LandmarkingParameters.PARAMETER_MIN));
                normalizer.setParameter(Normalization.PARAMETER_MAX, parameters.get(LandmarkingParameters.PARAMETER_MAX));

                normalizer.createPreprocessingModel(exampleSet);
                ExampleSet normalizedExampleSet = normalizer.doWork(exampleSet);

                return normalizedExampleSet;

            }

        return exampleSet;
    }

    /**
     * Returns the performance criterion average value specified.
     * 
     * @param criterion
     *            the name of the criterion to be returned.
     * @return The makro average if it was defined and the mikro average (the
     *         current value) otherwise.
     */
    protected Double getPerformanceCriterion(String criterion) {
        return performance.getCriterion(criterion).getAverage();
    }

    /**
     * @return a {@link PerformanceEvaluator} with accuracy and root mean
     *         squared error criteria set to true.
     * 
     */
    private Operator createPerformanceEvaluator() {
        Operator performanceEvaluator = null;

        try {
            performanceEvaluator = OperatorService.createOperator(PolynominalClassificationPerformanceEvaluator.class);

            performanceEvaluator.setParameter(PARAMETER_PERFORMANCE_ACCURACY, "true");

        } catch (OperatorCreationException e) {
            e.printStackTrace();
        }

        return performanceEvaluator;

    }

    /**
     * @return a {@link List} of {@link Double} values of the performance
     *         criteria: accuracy and rootmean squared error
     */
    public List<Double> getPerformanceValues() {
        List<Double> values = new ArrayList<Double>();
        values.add(getPerformanceCriterion(PARAMETER_PERFORMANCE_ACCURACY));

        return values;
    }

    /**
     * @return a {@link Double} representing the accuracy of the landmarker.
     */

    public Double getAccuracy() {
        return getPerformanceCriterion(PARAMETER_PERFORMANCE_ACCURACY);
    }

    /**
     * @return the {@link PerformanceVector} of the landmark.
     */
    public PerformanceVector getPerformanceVector() {
        return this.performance;
    }

    /**
     * @return the name of the landmark.
     */
    public String getName() {
        return name;
    }

}
