/*
 *  RapidMiner PaREn Automatic System Construction Extension 
 *
 *  Copyright (C) 2009-2010 by Deutsches Forschungszentrum fuer
 *  Kuenstliche Intelligenz GmbH or its licensors, as applicable.
 *  
 *  This is free software: you can redistribute it and/or modify 
 *  it under the terms of the GNU Affero General Public License as published by 
 *  the Free Software Foundation, either version 3 of the License, or 
 *  (at your option) any later version. 
 *  
 *  You should have received a copy of the GNU Affero General Public License 
 *  along with this software. If not, see <http://www.gnu.org/licenses/. 
 */
package de.dfki.madm.paren.landmarking;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.rapidminer.example.Attribute;
import com.rapidminer.example.ExampleSet;
import com.rapidminer.example.ExampleSetFactory;
import com.rapidminer.example.set.SplittedExampleSet;
import com.rapidminer.operator.Operator;
import com.rapidminer.operator.OperatorDescription;
import com.rapidminer.operator.OperatorException;
import com.rapidminer.operator.ports.InputPort;
import com.rapidminer.operator.ports.OutputPort;
import com.rapidminer.operator.ports.metadata.MetaData;
import com.rapidminer.operator.ports.metadata.SimplePrecondition;
import com.rapidminer.parameter.ParameterType;
import com.rapidminer.parameter.ParameterTypeBoolean;
import com.rapidminer.parameter.ParameterTypeCategory;
import com.rapidminer.parameter.ParameterTypeDouble;
import com.rapidminer.parameter.ParameterTypeInt;
import com.rapidminer.parameter.UndefinedParameterError;
import com.rapidminer.parameter.conditions.BooleanParameterCondition;
import com.rapidminer.parameter.conditions.EqualTypeCondition;

/**
 * This {@link Operator} takes an {@link ExampleSet} as input and delivers an
 * {@link ExampleSet} containing the performance values of different landmark
 * learners.
 * 
 * <p>
 * By default the following landmark learners are evaluated: Naive Bayes,
 * 1-Nearest Neighbor, Linear Discriminant, Best Stump Node, Average Node,
 * Randomly chosen Node, Worst Stump Node.
 * 
 * The learners can be evaluated using Cross Validation. By default a 10 fold
 * stratified cross-validation is performed.
 * </p>
 * 
 * @author Sarah Daniel
 * 
 */
public class LandmarkingOperator extends Operator {
    
    private InputPort exampleSetInput = getInputPorts().createPort("exampleset");
    private OutputPort landmarkingOutput = getOutputPorts().createPort("exampleset");

    public LandmarkingOperator(OperatorDescription description) {
        super(description);

        exampleSetInput.addPrecondition(new SimplePrecondition(exampleSetInput, new MetaData(ExampleSet.class)));

    }

    @Override
    public void doWork() throws OperatorException {

        ExampleSet exampleSet = exampleSetInput.getData();
        ExampleSet resultSet;
        try {
            resultSet = doLandmarking(exampleSet);
        } catch (Exception e){
            String message = "";
            if (e.getMessage() != null)
                message = e.getMessage();
            if (e.getCause() != null){
                message = message + " "+ e.getCause().getMessage();
            }
            throw new OperatorException("Fail during Landmarking: " + message , e);
        }
        landmarkingOutput.deliver(resultSet);
    }

    public ExampleSet doLandmarking(ExampleSet data) throws SecurityException, IllegalArgumentException, NoSuchMethodException, InstantiationException, IllegalAccessException, InvocationTargetException {
        LandmarkingResults landmarkingResults = evaluateLandmarkers(data);
        ExampleSet resultSet = initiateResultSet(landmarkingResults);
        return resultSet;
    }

    /**
     * @param exampleSet
     *            The {@link ExampleSet} to be evaluated.
     * @return A {@link LandmarkingResults} containing the results of the
     *         selected landmark learners.
     * @throws InvocationTargetException 
     * @throws IllegalAccessException 
     * @throws InstantiationException 
     * @throws NoSuchMethodException 
     * @throws IllegalArgumentException 
     * @throws SecurityException 
     */
    private LandmarkingResults evaluateLandmarkers(ExampleSet exampleSet) throws SecurityException, IllegalArgumentException, NoSuchMethodException, InstantiationException, IllegalAccessException, InvocationTargetException {
        try {

            if (exampleSet.size() == 0) {
                System.out.println("ExampleSet is empty....");
                return null;
            }

            LandmarkingResults landmarkersResult = new LandmarkingResults();
            Map<String, String> parameters = new HashMap<String, String>();
            addParameters(parameters);

            evaluateLandmarker(DecisionNodeLm.class, LandmarkingParameters.PARAMETER_DECISION_NODE_LM, exampleSet, landmarkersResult, parameters);
            evaluateLandmarker(AverageNodeLm.class, LandmarkingParameters.PARAMETER_AVERAGE_NODE_LM, exampleSet, landmarkersResult, parameters);
            evaluateLandmarker(RandomlyChosenNodeLm.class, LandmarkingParameters.PARAMETER_RANDOMLY_CHOSEN_NODE_LM, exampleSet, landmarkersResult, parameters);
            evaluateLandmarker(WorstNodeLm.class, LandmarkingParameters.PARAMETER_WORST_NODE_LM, exampleSet, landmarkersResult, parameters);
            evaluateLandmarker(NaiveBayesLm.class, LandmarkingParameters.PARAMETER_NAIVEBAYES_LM, exampleSet, landmarkersResult, parameters);
            evaluateLandmarker(LinearDiscriminantLm.class, LandmarkingParameters.PARAMETER_LINEAR_DISCRIMINANT_LM, exampleSet, landmarkersResult, parameters);
            parameters.put(LandmarkingParameters.PARAMETER_KNN_K_VALUE, getParameterAsString(LandmarkingParameters.PARAMETER_KNN_K_VALUE));
            evaluateLandmarker(KNNLearnerLm.class, LandmarkingParameters.PARAMETER_KNN_LM, exampleSet, landmarkersResult, parameters);

            return landmarkersResult;
        } catch (UndefinedParameterError e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * Adds the parameters with their values in the {@link Map} parameters.
     * 
     * @param parameters
     *            A {@link Map}, where the parameters should be added.
     */
    private void addParameters(Map<String, String> parameters) {
        try {
            // cross validation parameters
            parameters.put(LandmarkingParameters.PARAMETER_CROSS_VALIDATION, getParameterAsString(LandmarkingParameters.PARAMETER_CROSS_VALIDATION));
            parameters.put(LandmarkingParameters.PARAMETER_LEAVE_ONE_OUT, getParameterAsString(LandmarkingParameters.PARAMETER_LEAVE_ONE_OUT));
            parameters.put(LandmarkingParameters.PARAMETER_NUMBER_OF_VALIDATIONS, getParameterAsString(LandmarkingParameters.PARAMETER_NUMBER_OF_VALIDATIONS));
            parameters.put(LandmarkingParameters.PARAMETER_SAMPLING_TYPE, getParameterAsString(LandmarkingParameters.PARAMETER_SAMPLING_TYPE));
            parameters.put(LandmarkingParameters.PARAMETER_AVERAGE_PERFORMANCES_ONLY, getParameterAsString(LandmarkingParameters.PARAMETER_AVERAGE_PERFORMANCES_ONLY));

            // normalization parameters
            parameters.put(LandmarkingParameters.PARAMETER_NORMALIZE_EXAMPLESET, getParameterAsString(LandmarkingParameters.PARAMETER_NORMALIZE_EXAMPLESET));
            parameters.put(LandmarkingParameters.PARAMETER_NORMALIZATION_METHOD, getParameterAsString(LandmarkingParameters.PARAMETER_NORMALIZATION_METHOD));
            parameters.put(LandmarkingParameters.PARAMETER_MIN, getParameterAsString(LandmarkingParameters.PARAMETER_MIN));
            parameters.put(LandmarkingParameters.PARAMETER_MAX, getParameterAsString(LandmarkingParameters.PARAMETER_MAX));

        } catch (UndefinedParameterError e) {
            e.printStackTrace();
        }
    }

    /**
     * If the landmarkParameter is true, the landmarker is evaluated by creating
     * a new instance of the specified landmarkers name and adds its result to
     * the {@link LandmarkingResults} landmarkerResult.
     * 
     * @param clazz
     *            The {@link Class} of the Landmarker which has to be evaluated
     * @param landmarkParameter
     *            A {@link String} representing the parameter name, which
     *            indicates whether the landmarker should be included in the
     *            evaluation process or not.
     * @param exampleSet
     *            The {@link ExampleSet} to be evaluated by the landmarker.
     * @param landmarkersResult
     *            The {@link LandmarkingResults} object that should contain the
     *            result of the landmarkers.
     * @param parameters
     *            The parameters needed by the landmarker.
     * @throws NoSuchMethodException 
     * @throws SecurityException 
     * @throws InvocationTargetException 
     * @throws IllegalAccessException 
     * @throws InstantiationException 
     * @throws IllegalArgumentException 
     */
    private void evaluateLandmarker(Class<? extends AbstractLandmarker> clazz, String landmarkParameter, ExampleSet exampleSet, LandmarkingResults landmarkersResult, Map<String, String> parameters) throws SecurityException, NoSuchMethodException, IllegalArgumentException, InstantiationException, IllegalAccessException, InvocationTargetException {

            if (getParameterAsBoolean(landmarkParameter)) {

                Class<? extends AbstractLandmarker> landmarkerClass = clazz; // Class.forName(className);
                Constructor<? extends AbstractLandmarker> constructor = landmarkerClass.getDeclaredConstructor(new Class[] { com.rapidminer.example.ExampleSet.class, Map.class });
                AbstractLandmarker landmarker = constructor.newInstance(new Object[] { exampleSet, parameters });

                landmarkersResult.addLandMarker(landmarker.getName(), landmarker.getAccuracy());

            }

    }

    /**
     * @param landmarkingResults
     *            {@link LandmarkingIOObject} containing the results of the
     *            Landmark process.
     * @return {@link ExampleSet} containing the results of the landmark process
     *         obtained by the {@link LandmarkingResults}.
     */
    private ExampleSet initiateResultSet(LandmarkingResults landmarkingResults) {
        if (!(landmarkingResults.equals(null))) {
            Object[][] data = landmarkingResults.getPerformanceData();
            ExampleSet resultSet = ExampleSetFactory.createExampleSet(data);
            setAttributeNames(resultSet, landmarkingResults.landmarkerNames());
            return resultSet;
        }
        return null;
    }

    /**
     * Sets the attribute names of the {@link ExampleSet}.
     * 
     * @param exampleSet
     *            The {@link ExampleSet} containing the attribute names to be
     *            changed.
     * @param attributeNames
     *            The new attribute names.
     */
    private void setAttributeNames(ExampleSet exampleSet, Object[] attributeNames) {

        Iterator<Attribute> attributes = exampleSet.getAttributes().allAttributes();
        for (Object name : attributeNames) {
            attributes.next().setName((String) name);
        }
    }

    @Override
    public List<ParameterType> getParameterTypes() {
        List<ParameterType> types = super.getParameterTypes();

        // Parameters for the landmarkers
        types.add(new ParameterTypeBoolean(LandmarkingParameters.PARAMETER_NAIVEBAYES_LM, "If checked, the Naive Bayes landmarker will be evaluated ", true, true));
        types.add(new ParameterTypeBoolean(LandmarkingParameters.PARAMETER_KNN_LM, "If checked, the K-Nearest Neighbour will be included in the landmarking process", true, true));

        ParameterType type = new ParameterTypeInt(LandmarkingParameters.PARAMETER_KNN_K_VALUE, "The k value for the k-Nearest Neighbour", 1, Integer.MAX_VALUE, 1, true);
        type.registerDependencyCondition(new BooleanParameterCondition(this, LandmarkingParameters.PARAMETER_KNN_LM, true, true));
        types.add(type);

        types.add(new ParameterTypeBoolean(LandmarkingParameters.PARAMETER_DECISION_NODE_LM, "If checked, the best stump node landmarker will be evaluated", true, true));
        types.add(new ParameterTypeBoolean(LandmarkingParameters.PARAMETER_AVERAGE_NODE_LM, "If checked the average node landmarker will be evaluated.", true, true));
        types.add(new ParameterTypeBoolean(LandmarkingParameters.PARAMETER_RANDOMLY_CHOSEN_NODE_LM, "If checked, the randomly chosen node landmarker will be evaluated", true, true));
        types.add(new ParameterTypeBoolean(LandmarkingParameters.PARAMETER_WORST_NODE_LM, "If checked, the worst node landmarker will be evaluated", true, true));
        types.add(new ParameterTypeBoolean(LandmarkingParameters.PARAMETER_LINEAR_DISCRIMINANT_LM, "If checked, the linear iscrimanant landmarker will be evaluated", true, true));

        // Cross validation parameters
        types.add(new ParameterTypeBoolean(LandmarkingParameters.PARAMETER_CROSS_VALIDATION, "If checked all landmark learners will be evaluated using cross-validation.", false, true));
        type = new ParameterTypeBoolean(LandmarkingParameters.PARAMETER_LEAVE_ONE_OUT, "Set the number of validations to the number of examples. If set to true, number_of_validations is ignored",
                false, true);
        type.registerDependencyCondition(new BooleanParameterCondition(this, LandmarkingParameters.PARAMETER_CROSS_VALIDATION, true, true));
        types.add(type);

        type = new ParameterTypeInt(LandmarkingParameters.PARAMETER_NUMBER_OF_VALIDATIONS, "Number of subsets for the crossvalidation.", 2, Integer.MAX_VALUE, 10, true);
        type.registerDependencyCondition(new BooleanParameterCondition(this, LandmarkingParameters.PARAMETER_LEAVE_ONE_OUT, false, false));
        types.add(type);

        type = new ParameterTypeCategory(LandmarkingParameters.PARAMETER_SAMPLING_TYPE,
                "Defines the sampling type of the cross validation (linear = consecutive subsets, shuffled = random subsets, stratified = random subsets with class distribution kept constant)",
                SplittedExampleSet.SAMPLING_NAMES, SplittedExampleSet.STRATIFIED_SAMPLING);
        type.registerDependencyCondition(new BooleanParameterCondition(this, LandmarkingParameters.PARAMETER_LEAVE_ONE_OUT, false, false));
        types.add(type);

        type = new ParameterTypeBoolean(LandmarkingParameters.PARAMETER_AVERAGE_PERFORMANCES_ONLY,
                "Indicates if only performance vectors should be averaged or all types of averagable result vectors", true);
        type.registerDependencyCondition(new BooleanParameterCondition(this, LandmarkingParameters.PARAMETER_CROSS_VALIDATION, true, true));
        types.add(type);

        // Normalization Parameters
        types.add(new ParameterTypeBoolean(LandmarkingParameters.PARAMETER_NORMALIZE_EXAMPLESET, "Indicates if the dataset should be normalized or not, before being evaluated by the landmarker.",
                true, true));

        type = new ParameterTypeCategory(LandmarkingParameters.PARAMETER_NORMALIZATION_METHOD, "Select the normalization method.", LandmarkingParameters.NORMALIZATION_METHODS, 1);
        type.registerDependencyCondition(new BooleanParameterCondition(this, LandmarkingParameters.PARAMETER_NORMALIZE_EXAMPLESET, false, true));
        types.add(type);
        type = new ParameterTypeDouble(LandmarkingParameters.PARAMETER_MIN, "The minimum value after normalization", Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, 0.0d);
        type.registerDependencyCondition(new EqualTypeCondition(this, LandmarkingParameters.PARAMETER_NORMALIZATION_METHOD, LandmarkingParameters.NORMALIZATION_METHODS, true, new int[] { 1 }));
        types.add(type);
        type = new ParameterTypeDouble(LandmarkingParameters.PARAMETER_MAX, "The maximum value after normalization", Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY, 1.0d);
        type.registerDependencyCondition(new EqualTypeCondition(this, LandmarkingParameters.PARAMETER_NORMALIZATION_METHOD, LandmarkingParameters.NORMALIZATION_METHODS, true, new int[] { 1 }));
        types.add(type);

        return types;
    }

}
